#!usr/bin/env python
# -*- coding: utf-8 -*-
import platform
import argparse
from pip import main
from deprecated import deprecated
import click

from GANDLF import __version__
from GANDLF.entrypoints import append_copyright_to_help
from GANDLF.utils import get_git_hash
from GANDLF.utils import logger_setup

from GANDLF.cli import copyrightMessage


def _debug_info(verbose: bool):
    """Function to display necessary debugging information."""
    print(f"GANDLF version: {__version__}")
    print(f"Git hash: {get_git_hash()}")
    print(f"Platform: {platform.platform()}")
    print(f"Machine: {platform.machine()}")
    print(f"Processor: {platform.processor()}")
    print(f"Architecture: {(' ').join(list(platform.architecture()))}")
    print("Python environment:")
    print(f"  Version: {platform.python_version()}")
    print(f"  Implementation: {platform.python_implementation()}")
    print(f"  Compiler: {platform.python_compiler()}")
    print(f"  Build: {(' ').join(list(platform.python_build()))}")
    if verbose:
        print("  Installed packages:")
        print(main(["list"]))


@click.command()
@click.option(
    "--verbose",
    "-v",
    is_flag=True,
    help="If passed, prints all packages installed as well",
)
@click.option(
    "--log-file",
    type=click.Path(),
    default=None,
    help="Output file which will contain the logs.",
)
@append_copyright_to_help
def new_way(verbose: bool, log_file):
    """Displays detailed info about system environment: library versions, settings, etc."""

    logger_setup(log_file)
    _debug_info(verbose=verbose)


# main function
@deprecated(
    "This is a deprecated way of running GanDLF. Please, use `gandlf debug-info` cli command "
    + "instead of `gandlf_debugInfo`.\n"
    + "`gandlf_debugInfo` script would be deprecated soon."
)
def old_way():
    parser = argparse.ArgumentParser(
        prog="GANDLF_DebugInfo",
        formatter_class=argparse.RawTextHelpFormatter,
        description="Generate debugging information for maintainers.\n\n"
        + copyrightMessage,
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="If True, prints all packages installed as well.",
    )
    args = parser.parse_args()
    logger_setup()
    _debug_info(verbose=args.verbose)


if __name__ == "__main__":
    old_way()



# adapted from https://github.com/fepegar/torchio/blob/main/src/torchio/transforms/augmentation/intensity/random_blur.py

from collections import defaultdict
from typing import Union, Tuple

import torch

from torchio.typing import TypeTripletFloat, TypeSextetFloat
from torchio.data.subject import Subject
from torchio.transforms import IntensityTransform
from torchio.transforms.augmentation import RandomTransform
from torchio.transforms.augmentation.intensity.random_blur import Blur


class RandomBlurEnhanced(RandomTransform, IntensityTransform):
    r"""Blur an image using a random-sized Gaussian filter.

    Args:
        std: Tuple :math:`(a_1, b_1, a_2, b_2, a_3, b_3)` representing the
            ranges (in mm) of the standard deviations
            :math:`(\sigma_1, \sigma_2, \sigma_3)` of the Gaussian kernels used
            to blur the image along each axis, where
            :math:`\sigma_i \sim \mathcal{U}(a_i, b_i)`.
            If two values :math:`(a, b)` are provided,
            then :math:`\sigma_i \sim \mathcal{U}(a, b)`.
            If only one value :math:`x` is provided,
            then :math:`\sigma_i \sim \mathcal{U}(0, x)`.
            If three values :math:`(x_1, x_2, x_3)` are provided,
            then :math:`\sigma_i \sim \mathcal{U}(0, x_i)`.
        **kwargs: See :class:`~torchio.transforms.Transform` for additional
            keyword arguments.
    """

    def __init__(self, std: Union[float, Tuple[float, float]] = None, **kwargs):
        super().__init__(**kwargs)
        self.std_original = std

    def apply_transform(self, subject: Subject) -> Subject:
        arguments = defaultdict(dict)
        for name, image in self.get_images_dict(subject).items():
            self.std_ranges = self.calculate_std_ranges(image)
            arguments["std"][name] = self.get_params(self.std_ranges)
        transform = Blur(**self.add_include_exclude(arguments))
        transformed = transform(subject)
        return transformed

    def get_params(self, std_ranges: TypeSextetFloat) -> TypeTripletFloat:
        std = self.sample_uniform_sextet(std_ranges)
        return std

    def calculate_std_ranges(self, image: torch.Tensor) -> Tuple[float, float]:
        std_ranges = self.std_original
        if self.std_original is None:
            # calculate the default std range based on 1.5% of the input image std - https://github.com/mlcommons/GaNDLF/issues/518
            std_ranges = (0, 0.015 * torch.std(image.data.float()).item())
        return self.parse_params(std_ranges, None, "std", min_constraint=0)



from typing import Union
from pydantic import BaseModel, model_validator, Field, AfterValidator
from GANDLF.configuration.default_config import DefaultParameters
from GANDLF.configuration.differential_privacy_config import DifferentialPrivacyConfig
from GANDLF.configuration.nested_training_config import NestedTraining
from GANDLF.configuration.optimizer_config import OptimizerConfig
from GANDLF.configuration.patch_sampler_config import PatchSamplerConfig
from GANDLF.configuration.scheduler_config import SchedulerConfig
from GANDLF.utils import version_check
from importlib.metadata import version
from typing_extensions import Self, Literal, Annotated
from GANDLF.configuration.validators import (
    validate_scheduler,
    validate_optimizer,
    validate_loss_function,
    validate_metrics,
    validate_data_preprocessing,
    validate_patch_size,
    validate_parallel_compute_command,
    validate_patch_sampler,
    validate_data_augmentation,
    validate_data_postprocessing_after_reverse_one_hot_encoding,
    validate_differential_privacy,
)
from GANDLF.configuration.model_config import ModelConfig


class Version(BaseModel):
    minimum: str
    maximum: str

    @model_validator(mode="after")
    def validate_version(self) -> Self:
        if version_check(self.model_dump(), version_to_check=version("GANDLF")):
            return self


class InferenceMechanismConfig(BaseModel):
    grid_aggregator_overlap: Literal["crop", "average"] = Field(default="crop")
    patch_overlap: int = Field(default=0)


class UserDefinedParameters(DefaultParameters):
    version: Version = Field(
        default=Version(minimum=version("GANDLF"), maximum=version("GANDLF")),
        description="GANDLF version",
    )
    patch_size: Union[list[Union[int, float]], int, float] = Field(
        description="Patch size."
    )
    model: ModelConfig = Field(description="The model to use. ")
    modality: Literal["rad", "histo", "path"] = Field(description="Modality.")
    loss_function: Annotated[
        Union[dict, str],
        Field(description="Loss function."),
        AfterValidator(validate_loss_function),
    ]
    metrics: Annotated[
        Union[dict, list[Union[str, dict, set]]],
        Field(description="Metrics."),
        AfterValidator(validate_metrics),
    ]
    nested_training: NestedTraining = Field(description="Nested training.")
    parallel_compute_command: str = Field(
        default="", description="Parallel compute command."
    )
    scheduler: Union[str, SchedulerConfig] = Field(
        description="Scheduler.", default=SchedulerConfig(type="triangle_modified")
    )
    optimizer: Union[str, OptimizerConfig] = Field(
        description="Optimizer.", default=OptimizerConfig(type="adam")
    )
    patch_sampler: Union[str, PatchSamplerConfig] = Field(
        description="Patch sampler.", default=PatchSamplerConfig()
    )
    inference_mechanism: InferenceMechanismConfig = Field(
        description="Inference mechanism.", default=InferenceMechanismConfig()
    )
    data_postprocessing_after_reverse_one_hot_encoding: dict = Field(
        description="data_postprocessing_after_reverse_one_hot_encoding.", default={}
    )
    differential_privacy: Union[bool, DifferentialPrivacyConfig] = Field(
        description="Differential privacy.", default=None
    )
    clip_mode: Literal["norm", "value"] = Field(
        description="Clip mode.", default="norm"
    )
    data_preprocessing: Annotated[
        dict,
        Field(description="Data preprocessing."),
        AfterValidator(validate_data_preprocessing),
    ] = {}
    data_augmentation: Annotated[dict, Field(description="Data augmentation.")] = {}

    # Validators
    @model_validator(mode="after")
    def validate(self) -> Self:
        # validate the patch_size
        self.patch_size, self.model.dimension = validate_patch_size(
            self.patch_size, self.model.dimension
        )
        # validate the parallel_compute_command
        self.parallel_compute_command = validate_parallel_compute_command(
            self.parallel_compute_command
        )
        # validate scheduler
        self.scheduler = validate_scheduler(
            self.scheduler, self.learning_rate, self.num_epochs
        )
        # validate optimizer
        self.optimizer = validate_optimizer(self.optimizer)
        # validate patch_sampler
        self.patch_sampler = validate_patch_sampler(self.patch_sampler)
        # validate_data_augmentation
        self.data_augmentation = validate_data_augmentation(
            self.data_augmentation, self.patch_size
        )
        # validate data_postprocessing_after_reverse_one_hot_encoding
        (
            self.data_postprocessing_after_reverse_one_hot_encoding,
            self.data_postprocessing,
        ) = validate_data_postprocessing_after_reverse_one_hot_encoding(
            self.data_postprocessing_after_reverse_one_hot_encoding,
            self.data_postprocessing,
        )
        # validate differential_privacy
        self.differential_privacy = validate_differential_privacy(
            self.differential_privacy, self.batch_size
        )

        return self
